"""
Utilities used by crank.

Copyright (c) Chrispther Perkins
MIT License
"""

import collections, sys, string, inspect
import warnings

__all__ = [
        'get_argspec', 'get_params_with_argspec', 'remove_argspec_params_from_params',
        'method_matches_args', 'Path', 'default_path_translator', 'flatten_arguments'
    ]


_PY2 = bool(sys.version_info[0] == 2)


class _NotFound(object):
    pass


def _getargspec(func):
    if not hasattr(inspect, 'signature'):
        return inspect.getargspec(func)
    else:  #pragma: no cover
        sig = inspect.signature(func)
        args = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        ]
        varargs = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.VAR_POSITIONAL
        ]
        varargs = varargs[0] if varargs else None
        varkw = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.VAR_KEYWORD
        ]
        varkw = varkw[0] if varkw else None
        defaults = tuple((
            p.default for p in sig.parameters.values()
            if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD and p.default is not p.empty
        )) or None
        return args, varargs, varkw, defaults


_cached_argspecs = {}
def get_argspec(func):
    if _PY2:
        im_func = getattr(func, 'im_func', func)
    else:  #pragma: no cover
        im_func = getattr(func, '__func__', func)

    try:
        argspec = _cached_argspecs[im_func]
    except KeyError:
        spec = _getargspec(im_func)
        argvals = spec[3]

        # this is a work around for a crappy api choice in getargspec
        if argvals is None:
            argvals = []

        argspec = _cached_argspecs[im_func] = (spec[0][1:], spec[1], spec[2], argvals)

    return argspec


def get_params_with_argspec(func, params, remainder):
    argvars, var_args, argkws, argvals = get_argspec(func)

    if argvars and remainder:
        params = params.copy()
        remainder_len = len(remainder)
        for i, var in enumerate(argvars):
            if i >= remainder_len:
                break
            params[var] = remainder[i]
    return params


def remove_argspec_params_from_params(func, params, remainder):
    """Remove parameters from the argument list that are
       not named parameters
       Returns: params, remainder"""

    warnings.warn("remove_argspec_params_from_params is deprecated and will be removed",
                  DeprecationWarning, stacklevel=2)

    # figure out which of the vars in the argspec are required
    argvars, var_args, argkws, argvals = get_argspec(func)

    # if there are no required variables, or the remainder is none, we
    # have nothing to do
    if not argvars or not remainder:
        return params, remainder

    required_vars = argvars
    optional_vars = []
    if argvals:
        required_vars = argvars[:-len(argvals)]
        optional_vars = argvars[-len(argvals):]

    # make a copy of the params so that we don't modify the existing one
    params = params.copy()

    # replace the existing required variables with the values that come in
    # from params. these could be the parameters that come off of validation.
    remainder = list(remainder)
    remainder_len = len(remainder)
    for i, var in enumerate(required_vars):
        val = params.get(var, _NotFound)
        if val is not _NotFound:
            if i < remainder_len:
                remainder[i] = val
            else:
                remainder.append(val)
            del params[var]

    # remove the optional positional variables (remainder) from the named parameters
    # until we run out of remainder, that is, avoid creating duplicate parameters
    for i, (original, var) in enumerate(zip(remainder[len(required_vars):],optional_vars)):
        if var in params:
            remainder[ len(required_vars)+i ] = params[var]
            del params[var]

    return params, tuple(remainder)


def flatten_arguments(func, params, remainder, keep_unexpected=False):
    """Returns all the arguments for a function as positional parameters.

    Keyword arguments are returned only if the function supports **kwargs
    """
    if remainder is None:
        remainder = tuple()

    # figure out which of the vars in the argspec are required
    positional_args, varargs, argkws, default_arg_values = get_argspec(func)

    if not params:
        if varargs:
            # If all arguments are already positional and we accept variable arguments
            # we have nothing to do, params are already flattened and there are no
            # extra arguments
            return tuple(remainder), params
        else:
            # Otherwise arguments are already positional, but there are extra arguments
            # so we just throw away the extra arguments
            return tuple(remainder[:len(positional_args)]), params

    args = []
    kwargs = params.copy()

    # Gather positional arguments
    for idx, argname in enumerate(positional_args):
        val = kwargs.pop(argname, _NotFound)
        if val is not _NotFound:
            args.append(val)
        elif idx < len(remainder):
            args.append(remainder[idx])
        else:
            # if argument is not available look for default value or just stop
            # as we are actually missing an argument
            try:
                default_arg_idx = idx - (len(positional_args) - len(default_arg_values))
                if default_arg_idx < 0:
                    raise IndexError()
                args.append(default_arg_values[default_arg_idx])
            except IndexError:
                raise TypeError('{0} missing "{1}" required argument'.format(func, argname))

    if varargs:
        args.extend(remainder[len(args):])

    if not argkws:
        kwargs = {}

    return tuple(args), kwargs


def method_matches_args(method, params, remainder, lax_params=False):
    """
    This method matches the params from the request along with the remainder to the
    method's function signiture.  If the two jive, it returns true.

    It is very likely that this method would go into ObjectDispatch in the future.
    """
    argvars, ovar_args, argkws, argvals = get_argspec(method)

    required_vars = argvars
    if argvals:
        required_vars = argvars[:-len(argvals)]

    params = params.copy()

    #remove the appropriate remainder quotient
    if len(remainder)<len(required_vars):
        #pull the first few off with the remainder
        required_vars = required_vars[len(remainder):]
    else:
        #there is more of a remainder than there is non optional vars
        required_vars = []

    #remove vars found in the params list
    for var in required_vars[:]:
        if var in params:
            required_vars.pop(0)
            # remove the param from the params so when we see if
            # there are params that arent in the non-required vars we
            # can evaluate properly
            del params[var]
        else:
            break

    #remove params that have a default value
    vars_with_default = argvars[len(argvars)-len(argvals):]
    for var in vars_with_default:
        if var in params:
            del params[var]

    #make sure no params exist if keyword argumnts are missing
    if not lax_params and argkws is None and params:
        return False

    #make sure all of the non-optional-vars are there
    if not required_vars:
        #there are more args in the remainder than are available in the argspec
        if len(argvars)<len(remainder) and not ovar_args:
            return False
        return True


    return False


if _PY2: #pragma: no cover
    translation_dict = dict([(ord(c), unicode('_')) for c in unicode(string.punctuation)])
    translation_string = string.maketrans(string.punctuation,
                                          '_' * len(string.punctuation))
else: #pragma: no cover
    translation_dict = None
    translation_string = str.maketrans(string.punctuation,
                                       '_' * len(string.punctuation))


def default_path_translator(path_piece):
    if isinstance(path_piece, str):
        return path_piece.translate(translation_string)
    else: #pragma: no cover
        return path_piece.translate(translation_dict)


def noop_translation(path_piece):
    return path_piece


class Path(collections.deque):
    def __init__(self, value=None, separator='/'):
        self.separator = separator

        super(Path, self).__init__()

        if value is not None:
            self._assign(value)

    def _assign(self, value):
        separator = self.separator
        self.clear()

        if not _PY2: # pragma: no cover
            string_types = str
        else: # pragma: no cover
            string_types = basestring

        if isinstance(value, string_types):
            self.extend(value.split(separator))
            return

        self.extend(value)

    def __set__(self, obj, value):
        self._assign(value)

    def __str__(self):
        return str(self.separator).join(self)

    def __unicode__(self):  # pragma: no cover
        #unused on PY3
        return unicode(self.separator).join(self)

    def __repr__(self):
        return "<Path %r>" % super(Path, self).__repr__()

    def __eq__(self, other):
        return type(other)(self) == other

    def __getitem__(self, i):
        try:
            return super(Path, self).__getitem__(i)

        except TypeError:
            return Path([self[i] for i in range(*i.indices(len(self)))])



"""
This module implements the :class:`DispatchState` class
"""
import warnings

from crank.util import default_path_translator, noop_translation

try:
    string_type = basestring
except NameError: # pragma: no cover
    string_type = str


class DispatchState(object):
    """
    This class keeps around all the pertainent info for the state
    of the dispatch as it traverses through the tree.  This allows
    us to attach things like routing args and to keep track of the
    path the controller takes along the system.
    
    Arguments:
        request 
              object, must have a path_info attribute if path_info is not provided
        dispatcher
              dispatcher object to get the ball rolling
        params
              parameters to pass into the dispatch state will use request.params
        path_info
              pre-split list of path elements, will use request.pathinfo if not used
        strip_extension
              Whenever crank should strip the url extension or not resolving the path
        path_translator
              Function used to perform path escaping when looking for controller methods,
              can be None to perform no escaping or True to use default escaping function.
    """

    def __init__(self, request, dispatcher, params=None, path_info=None,
                 ignore_parameters=None, strip_extension=True, path_translator=None):
        self._request = request

        if path_translator is None:
            path_translator = noop_translation
        elif path_translator is True:
            path_translator = default_path_translator
        self._path_translator = path_translator

        self._strip_extension = strip_extension
        self.set_path(path_info)

        self._ignored_parameters = ignore_parameters
        if params is None:
            params = request.params
        self.set_params(params)

        self._root_dispatcher = dispatcher
        self._controller = None
        self._controller_path = []
        self._routing_args = {}
        self._action = None
        self._remainder = None
        self._notfound_stack = []

        self.add_controller('/', dispatcher)

    @property
    def root_dispatcher(self):
        """Root Dispatcher instance that initiated the dispatch flow"""
        return self._root_dispatcher

    @property
    def request(self):
        """The request that originated the dispatch process"""
        return self._request

    @property
    def path(self):
        """The path (URL) that has to be dispatched"""
        return self._path

    @property
    def extension(self):
        """Extension of the URL (only if strip_extension is enabled).

        If the path ends with an extension and strip_extension is enabled
        the extension is stripped from the url and is available here.
        """
        return self._extension

    @property
    def controller(self):
        """Controller currently handling the request"""
        return self._controller

    @property
    def controller_path(self):
        """Controllers that got traversed to dispatch the request"""
        return tuple(self._controller_path)

    @property
    def action(self):
        """Method in charge of processing the request"""
        return self._action

    @property
    def params(self):
        """Parameters passed to the action"""
        return self._params

    @property
    def remainder(self):
        """Part of the URL path remaining after lookup of the action.

        Those is usually passed as positional arguments to the method
        in charge of the action together with params.
        """
        return self._remainder

    def add_controller(self, location, controller):
        """Add a controller object to the stack"""
        self._controller = controller
        self._controller_path.append((location, controller))

    def set_action(self, method, remainder):
        """Add the final method that will be called in the _call method"""
        self._action = method
        self._remainder = remainder

    def set_params(self, params):
        """Set parameters that will be passed to the called action"""
        self._params = params
        if self._ignored_parameters:
            for param in self._ignored_parameters:
                self._params.pop(param, None)

    def set_path(self, path_info):
        """Update path that needs to be dispatched.

        In case this is changed during the dispatch process it won't
        have any effect on the dispatch currently under execution.
        """
        path = path_info
        if path is None:
            path = self._request.path_info[1:]
            path = path.split('/')
        elif isinstance(path, string_type):
            path = path.split('/')

        try:
            if not path[0]:
                path = path[1:]
        except IndexError:
            pass

        try:
            while not path[-1]:
                path = path[:-1]
        except IndexError:
            pass

        # rob the extension
        self._extension = None
        if self._strip_extension and len(path) > 0 and '.' in path[-1]:
            end = path[-1]
            end, ext = end.rsplit('.', 1)
            self._extension = ext
            path[-1] = end
        self._path = path

    def resolve(self):
        """Once a DispatchState is created resolving it performs the dispatch.

        Returns the updated DispatchState where ``.controller``, ``.action``,
        ``.params`` and ``.remainder`` properties all point to the controller
        and method that should process the request and to the method arguments.
        """
        if self._action is not None:
            raise RuntimeError('Trying to resolve an already resolved DispatchState')
        return self._root_dispatcher._dispatch(self, self._path)

    def translate_path_piece(self, path_piece):
        return self._path_translator(path_piece=path_piece)

    def add_routing_args(self, current_path, remainder, fixed_args, var_args):
        """
        Add the "intermediate" routing args for a given controller mounted
        at the current_path. This is mostly used during REST dispatch to keep
        track of intermediate arguments and make them always available.
        """
        i = 0
        for i, arg in enumerate(fixed_args):
            if i >= len(remainder):
                break
            self._routing_args[arg] = remainder[i]
        remainder = remainder[i:]
        if var_args and remainder:
            self._routing_args[current_path] = remainder

    @property
    def routing_args(self):
        """Parameters detected by the routing system.

        This includes Request parameters and parameters extracted in other
        ways (usually added through :meth:`.add_routing_args`). In case
        of REST it will include intermediate arguments retrieved during dispatch
        of parent controllers.
        """
        return self._routing_args

    def add_method(self, method, remainder):  # pragma: no cover
        warnings.warn("add_method is deprecated, please use set_action instead",
                      DeprecationWarning, stacklevel=2)
        self.set_action(method, remainder)

    @property
    def method(self):  # pragma: no cover
        warnings.warn(".method is deprecated, please use .action instead",
                      DeprecationWarning, stacklevel=2)
        return self.action

    @property
    def dispatcher(self):  # pragma: no cover
        warnings.warn(".dispatcher is deprecated, please use .root_dispatcher instead",
                      DeprecationWarning, stacklevel=2)
        return self.root_dispatcher



from setuptools import setup, find_packages
import sys, os

version = '0.8.0'

setup(name='crank',
      version=version,
      description="Generalization of dispatch mechanism for use across frameworks.",
      long_description="""\
""",
      classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python',
      ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Christopher Perkins',
      author_email='chris@percious.com',
      url='https://github.com/TurboGears/crank',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      extras_require={
           # Used by Travis and Coverage due to setup.py nosetests
           # causing a coredump when used with coverage
           'testing': ['webob'],
      },
      install_requires=[
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
